### Remotely connected traversal source

The _DSE Java driver_ also provides full compatibility with [Apache TinkerPop][tinkerpop]'s 
query execution model and result types.

It is fully compatible with Tinkerpop's concept of remote traversal sources,
and may be used as an alternative to [building queries with the Fluent API](../querying/fluent_api) 
however, it may offer fewer features than the Fluent API queries would be able to offer in 
the future.

The module with the classes necessary to be able to use remote connected traversal sources 
with DSE Graph can be found on Maven:

```xml
<dependency>
    <groupId>com.datastax.dse</groupId>
    <artifactId>dse-java-driver-graph</artifactId>
    <version>1.8.0</version>
</dependency>
```

The following is an example of how to get a [Apache TinkerPop][tinkerpop] [GraphTraversalSource][GraphTraversalSource] 
that is remotely connected to a [DseGraph][DseGraph] server, and which communicates to 
[DseGraph][DseGraph] via the _DSE Java driver_ under the covers:

```java
DseCluster dseCluster = DseCluster.builder()
    .addContactPoint("1.2.3.4")
    .withGraphOptions(new GraphOptions()
                        .setGraphName("mygraph"))
    .build();
DseSession dseSession = dseCluster.connect();

GraphTraversalSource g = DseGraph.traversal(dseSession);

// Now you can use the Traversal source and use it **as if** it was working against a local graph, and with the usual TinkerPop API. All the communication with the DSE Graph server is done transparently.
List<Vertex> vertices = g.V().hasLabel("person").toList();
```

Traversal sources with different configurations can easily be created. By default the options
specific to [DseGraph][DseGraph] are taken from the [DseCluster][DseCluster] configuration, however
the API exposes a way to override each individual setting, per traversal source:

```java
GraphTraversalSource gOLTP = DseGraph.traversal(dseSession, new GraphOptions().setGraphName("mygraph"));
GraphTraversalSource gOLAP = DseGraph.traversal(dseSession, new GraphOptions().setGraphName("myothergraph").setGraphSource(ANALYTICS_SOURCE_NAME));

Vertex v = gOLTP.V().has("name", "marko").next();
long count = gOLAP.V().count().next();
```

**Important note**: there is no interactivity with DSE Graph until a [Terminal Step][TerminalStep] 
(such as `next()`, `toList()`, etc.) is called.

#### A word on Results

Objects returned after the Remote traversal execution are objects that are detached from
the original DSE Graph on the server. A _detached_ element is an element that lives on
its own, even though the _detached_ elements contains the complete data, modifications
made to a _detached_ element **do not affect** the data stored in the _DSE Graph_.

#### Additional features

Just like when [building queries with the Fluent API](../querying/fluent_api), using the 
DSE Java driver allows to use some advanced features from Tinkerpop. For instance one would 
use the Gremlin Domain Specific Language feature with a remote connected traversal source 
like this:

```java
// see TinkerPop documentation link for the generation of SocialTraversalSource
SocialTraversalSource gSocial = DseGraph.traversal(dseSession, SocialTraversalSource.class);
List<Vertex> vertices = gSocial.persons("marko").knows("vadas").toList();
```

DSE Search Text and Geo predicates can also be used with a remote connected traversal source:

```java
GraphTraversalSource g = DseGraph.traversal(dseSession, new GraphOptions().setGraphName("thegraph"));
Vertex v = g.V().has("textProp", Search.tokenPrefix("1")).next();
``` 

See [this page](../querying/fluent_api#dse-search-text-and-geo-gremlin-predicates) for more 
information.


[tinkerpop]: http://tinkerpop.apache.org/
[DseCluster]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/DseCluster.html
[GraphTraversalSource]: https://tinkerpop.apache.org/javadocs/3.3.0/full/org/apache/tinkerpop/gremlin/process/traversal/dsl/graph/GraphTraversalSource.html
[TerminalStep]: http://tinkerpop.apache.org/docs/current/reference/#terminal-steps
[Traversal]: https://tinkerpop.apache.org/javadocs/3.3.0/full/org/apache/tinkerpop/gremlin/process/traversal/Traversal.html
[DseGraph]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/dse/graph/api/DseGraph.html
