### Handling results

Graph queries return a [GraphResultSet], which is essentially an iterable of [GraphNode]:

```java
GraphStatement graphStatement = DseGraph.statementFromTraversal(DseGraph.traversal().V());
GraphResultSet rs = dseSession.executeGraph(graphStatement);

// Iterating:
for (GraphNode n : rs) {
    System.out.println(n);
}

// Get the first result only (or if you know there is exactly one):
GraphNode n = rs.one();
```

`GraphNode` wraps the responses returned by the server. You can coerce the result to a specific type using the
[asXxx()][Node] methods:

```java
GraphStatement graphStatement = DseGraph.statementFromTraversal(
        DseGraph.traversal().
        V().
        count());
GraphNode n = dseSession.executeGraph(graphStatement).one();
System.out.printf("The graph has %s vertices%n", n.asInt());
```

If the result is an array or an object (non-leaf node), you can iterate its child elements:

```java
if (n.isArray()) {
    for (int i = 0; i < n.size(); i++) {
        GraphNode child = n.get(i);
        System.out.printf("Element at position %d: %s%n", i, child);
    }
}

if (n.isObject()) {
    Iterator<String> fieldNames = n.fieldNames();
    while (fieldNames.hasNext()) {
        String fieldName = fieldNames.next();
        System.out.printf("Element at key %s: %s%n", fieldName, n.get(fieldName));
    }
}
```

The driver also exposes general-purpose methods to handle results in the form of _Maps_ and _Lists_:

```java
GraphStatement graphStatement = DseGraph.statementFromTraversal(
        DseGraph.traversal().
        V().
        valueMap());
GraphNode n = dseSession.executeGraph(graphStatement).one();
Map<String, Object> values = n.asMap();
```

#### Graph structural types

The driver has client-side representations for [Vertex], [Edge], [Path], [VertexProperty], and [Property].

These are accessible via the corresponding [GraphNode#asXXXX()][GraphNode] methods:

```java

GraphTraversalSource g = DseGraph.traversal();

GraphNode n = dseSession.executeGraph(DseGraph.statementFromTraversal(
        g.V().hasLabel("test_vertex")
    )).one();
Vertex vertex = n.asVertex();

n = dseSession.executeGraph(DseGraph.statementFromTraversal(
        g.V().hasLabel("test_vertex").
        outE()
    )).one();
Edge edge = n.asEdge();

n = dseSession.executeGraph(DseGraph.statementFromTraversal(
        g.V().hasLabel("test_vertex").
        outE().
        inV().
        path()
    )).one();

Path path = n.asPath();

n = dseSession.executeGraph(DseGraph.statementFromTraversal(
        g.V().hasLabel("test_vertex").
        properties("name")
    )).one();
// .properties() returns a list of properties, so we get the first one and transform it as a VertexProperty
VertexProperty vertexProperty = n.ge(0).asVertexProperty();
```

#### Data types compatibility matrix

_DSE Graph_ exposes several [data types](http://docs.datastax.com/en/latest-dse/datastax_enterprise/graph/reference/refDSEGraphDataTypes.html)
when defining a schema for a graph.

Those data types server-side translate into specific Java classes when the data is returned from the server.

Here is the exhaustive list of possible _DSE Graph_ data types, and their corresponding classes
in the DSE Java driver:

<table border="1" width="100%">
<tr><th>DSE Graph</th><th>Java Driver</th></tr>
<tr><td>bigint</td><td><tt>Long</tt></td></tr>
<tr><td>int</td><td><tt>Integer</tt></td></tr>
<tr><td>double</td><td><tt>Double</tt></td></tr>
<tr><td>float</td><td><tt>Float</tt></td></tr>
<tr><td>uuid</td><td><tt>UUID</tt></td></tr>
<tr><td>bigdecimal</td><td><tt>BigDecimal</tt></td></tr>
<tr><td>duration</td><td><tt>java.time.Duration</tt></td></tr>
<tr><td>inet</td><td><tt>InetAddress</tt></td></tr>
<tr><td>timestamp</td><td><tt>java.time.Instant</tt></td></tr>
<tr><td>time</td><td><tt>java.time.LocalTime</tt></td></tr>
<tr><td>date</td><td><tt>java.time.LocalDate</tt></td></tr>
<tr><td>smallint</td><td><tt>Short</tt></td></tr>
<tr><td>varint</td><td><tt>BigInteger</tt></td></tr>
<tr><td>polygon</td><td><tt>Polygon</tt></td></tr>
<tr><td>point</td><td><tt>Point</tt></td></tr>
<tr><td>linestring</td><td><tt>LineString</tt></td></tr>
<tr><td>blob</td><td><tt>byte[]</tt></td></tr>
</table>

#### Deserializing complex data types

The driver exposes methods to deserialize data into more complex data types, 
as long as the server-side data type associated corresponds. Doing so requires using
the [`GraphNode#as(Class<T> clazz)`][GraphNode] method:

```java
GraphTraversalSource g = DseGraph.traversal();

GraphNode n = dseSession.executeGraph(DseGraph.statementFromTraversal(
        g.V().hasLabel("test_vertex")
    )).one();
Vertex vertex = n.asVertex();
UUID uuidProp = vertex.getProperty("uuidProp").getValue().as(UUID.class);
```

#### Java 8 Time types

The _DSE Java driver_ is compatible with Java from version 6. The driver is able to 
automatically determine whether the application it's used with is running on such a legacy Java version,
and will be able to deserialize objects differently accordingly.

If using the driver with Java version 8 or higher, `java.time` types will be usable when 
retrieving the data from a Traversal (see the types matrix above).

If the driver is used with an inferior version of Java, other classes will be usable when 
retrieving the data sent by the server. For Java versions < 8, the `Duration()` and `Time()` DSE Graph types will be exposed as Strings.
`Timestamp()` will be exposed as a `java.util.Date`, and `Date()` as a [com.datastax.driver.core.LocalDate][localdate].

#### A word on Properties

The vertex properties exposed by the driver (_VertexProperty_) respect the same behaviour as 
in Apache TinkerPop. First, a _VertexProperty_ is a property, with a simple value. But in
addition to that, a _VertexProperty_ can be the parent of a sublist of _Property_. 
If that's the case, the sub properties of _VertexProperty_ are called _meta-properties_.

Moreover, in a vertex the **same** property key can be associated to multiple vertex properties.
In this case, the vertex is said to have "multi-properties".

Here is the syntax for dealing with vertex properties with the _DataStax Java driver_ graph types:

```java
GraphTraversalSource g = DseGraph.traversal();

GraphNode n = dseSession.executeGraph(DseGraph.statementFromTraversal(
        g.V().hasLabel("test_vertex_props")
    )).one();
Vertex vertex = n.asVertex();

// there can be more than one VertexProperty with the key "multi_with_meta_props"
Iterator<VertexProperty> vertexProps = vertex.getProperties("multi_with_meta_props");

VertexProperty vertexProp1 = vertexProps.next();
// the value of the vertex property
int vertexProp1Value = vertexProp1.getValue().asInt();
// the meta-properties of the vertex property
Iterator<Property> metaProps1 = vertexProp1.getProperties();
Property metaProp11 = metaProps1.next();
double metaPropValue11 = metaProp11.getValue().asDouble(); 
Property metaProp12 = metaProps1.next();
double metaPropValue12 = metaProp12.getValue().asDouble(); 

// **multi-properties**.
VertexProperty vertexProp2 = vertexProps.next();
[...]
```

More on how to create and query multi-valued properties and meta-properties is in the
[Apache Tinkerpop documentation](http://tinkerpop.apache.org/docs/3.2.3/reference/#vertex-properties).


[executeGraphString]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/DseSession.html#executeGraph-java.lang.String-
[executeGraphStatement]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/DseSession.html#executeGraph-com.datastax.driver.dse.graph.GraphStatement-
[executeGraphAsync]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/DseSession.html#executeGraphAsync-com.datastax.driver.dse.graph.GraphStatement-
[Vertex]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/Vertex.html
[Edge]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/Edge.html
[Path]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/Path.html
[VertexProperty]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/VertexProperty.html
[Property]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/Property.html
[Node]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/serde/Node.html
[GraphNode]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/GraphNode.html
[GraphResultSet]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/dse/graph/GraphResultSet.html
[localdate]: http://docs.datastax.com/en/drivers/java-dse/1.8/com/datastax/driver/core/LocalDate.html